# Conversation of .c3d files to OpenSim marker.trc and ground reaction forces
# grf.mot for the Sinergia data set. This script can be used for other data sets
# as well, however, the column names and transformation conventions may be
# different. Also, note that here we do not distinguish between left and right
# foot, therefore the setup_grf.xml file has to be manually updated.
#
# author: Dimitar Stanev <jimstanev@gmail.com>
# contributors: Celine Provins, George Papoulias
##
import os
import opensim
import numpy as np
from utils import rotate_data_table, mm_to_m, plot_sto_file
from utils import lowess_bell_shape_kern, create_opensim_storage

##
# specify the .c3d file

static_file = 'static.c3d'
task_file = 'task.c3d'
input_dir = os.path.abspath('../experimental_data/')
c3d_dir = input_dir
output_dir = input_dir

# OpenSim data adapters
adapter = opensim.C3DFileAdapter()
trc_adapter = opensim.TRCFileAdapter()

##
# extract data for static trial

# get markers
static = adapter.read(os.path.join(c3d_dir, static_file), 1)
markers_static = static['markers']

# process markers_task
rotate_data_table(markers_static, [1, 0, 0], -90)
# rotate_data_table(markers_static, [0, 1, 0], 180)
# rotate_data_table(markers_static, [0, 1, 0], -90)
trc_adapter.write(markers_static, os.path.join(output_dir, 'static.trc'))

##
# extract data for task (e.g., walk, run)

# get markers and forces
task = adapter.read(os.path.join(c3d_dir, task_file), 1)
markers_task = task['markers']
forces_task = task['forces']

# process markers_task
rotate_data_table(markers_task, [1, 0, 0], -90)
# rotate_data_table(markers_task, [0, 1, 0], 180)
# rotate_data_table(markers_task, [0, 1, 0], -90)
trc_adapter = opensim.TRCFileAdapter()
trc_adapter.write(markers_task, os.path.join(output_dir, 'task.trc'))

# process forces
rotate_data_table(forces_task, [1, 0, 0], 90)
rotate_data_table(forces_task, [0, 1, 0], 180)
rotate_data_table(forces_task, [0, 0, 1], 180)
# mirror_z(forces_task, 'f1')
# mirror_z(forces_task, 'f2')

# conversion of unit (f -> N, p -> mm, tau -> Nmm)
mm_to_m(forces_task, 'p1')
mm_to_m(forces_task, 'p2')
mm_to_m(forces_task, 'm1')
mm_to_m(forces_task, 'm2')

# interpolate and fit splines to smooth the data
t = forces_task.getIndependentColumn()
labels = forces_task.getColumnLabels()
list_mat = list()
for label in labels:
    f = forces_task.getDependentColumn(label)
    list_mat.append(lowess_bell_shape_kern(t, f, output_dir=output_dir))

# construct the matrix of the forces (forces, moments, torques / right and left)
# (type opensim.Matrix)
forces_task_np = np.array(list_mat)
forces_task_mat = opensim.Matrix(len(t), 18)
for n in range(6):
    for j in range(3):
        for i in range(len(t)):
            forces_task_mat.set(i, 3 * n + j, forces_task_np[n, i, j])

# export forces
labels_list = ['ground_force_vx', 'ground_force_vy', 'ground_force_vz',
               'ground_force_px', 'ground_force_py', 'ground_force_pz',
               'ground_torque_x', 'ground_torque_y', 'ground_torque_z',
               '1_ground_force_vx', '1_ground_force_vy', '1_ground_force_vz',
               '1_ground_force_px', '1_ground_force_py', '1_ground_force_pz',
               '1_ground_torque_x', '1_ground_torque_y', '1_ground_torque_z']
force_sto = create_opensim_storage(t, forces_task_mat, labels_list)
force_sto.setName('GRF')
force_sto.printResult(force_sto, 'task_grf', output_dir, 0.001, '.mot')

# plot
plot_sto_file(os.path.join(output_dir, 'task.trc'),
              os.path.join(output_dir, 'task.pdf'), 3)
plot_sto_file(os.path.join(output_dir, 'task_grf.mot'),
              os.path.join(output_dir, 'task_grf.pdf'), 3)
